@@ -21,11 +21,12 @@
 
 __all__ = [ "Firewall" ]
 
-import os.path
+import os
 import sys
 import copy
 import time
 import traceback
+from typing import Dict, List
 from firewall import config
 from firewall import functions
 from firewall.core import ipXtables
@@ -45,6 +46,7 @@ from firewall.core.fw_helper import FirewallHelper
 from firewall.core.fw_policy import FirewallPolicy
 from firewall.core.fw_nm import nm_get_bus_name, nm_get_interfaces_in_zone
 from firewall.core.logger import log
+from firewall.core.io.io_object import IO_Object
 from firewall.core.io.firewalld_conf import firewalld_conf
 from firewall.core.io.direct import Direct
 from firewall.core.io.service import service_reader
@@ -54,6 +56,8 @@ from firewall.core.io.ipset import ipset_reader
 from firewall.core.ipset import IPSET_TYPES
 from firewall.core.io.helper import helper_reader
 from firewall.core.io.policy import policy_reader
+from firewall.core.io.functions import check_on_disk_config
+from firewall.core.rich import Rich_Rule
 from firewall import errors
 from firewall.errors import FirewallError
 
@@ -68,28 +72,12 @@ class Firewall(object):
         self._firewalld_conf = firewalld_conf(config.FIREWALLD_CONF)
         self._offline = offline
 
-        if self._offline:
-            self.ip4tables_enabled = False
-            self.ip6tables_enabled = False
-            self.ebtables_enabled = False
-            self.ipset_enabled = False
-            self.ipset_supported_types = IPSET_TYPES
-            self.nftables_enabled = False
-        else:
+        if not offline:
             self.ip4tables_backend = ipXtables.ip4tables(self)
-            self.ip4tables_enabled = True
-            self.ipv4_supported_icmp_types = [ ]
             self.ip6tables_backend = ipXtables.ip6tables(self)
-            self.ip6tables_enabled = True
-            self.ipv6_supported_icmp_types = [ ]
             self.ebtables_backend = ebtables.ebtables()
-            self.ebtables_enabled = True
             self.ipset_backend = ipset.ipset()
-            self.ipset_enabled = True
-            self.ipset_supported_types = IPSET_TYPES
             self.nftables_backend = nftables.nftables(self)
-            self.nftables_enabled = True
-
             self.modules_backend = modules.modules()
 
         self.icmptype = FirewallIcmpType(self)
@@ -116,7 +104,9 @@ class Firewall(object):
     def __init_vars(self):
         self._state = "INIT"
         self._panic = False
-        self._default_zone = ""
+        self._default_zone = config.FALLBACK_ZONE
+        self._default_zone_interfaces = []
+        self._nm_assigned_interfaces = []
         self._module_refcount = { }
         self._marks = [ ]
         # fallback settings will be overloaded by firewalld.conf
@@ -130,7 +120,77 @@ class Firewall(object):
         self._rfc3964_ipv4 = config.FALLBACK_RFC3964_IPV4
         self._allow_zone_drifting = config.FALLBACK_ALLOW_ZONE_DRIFTING
 
-    def _check_tables(self):
+        if self._offline:
+            self.ip4tables_enabled = False
+            self.ip6tables_enabled = False
+            self.ebtables_enabled = False
+            self.ipset_enabled = False
+            self.ipset_supported_types = IPSET_TYPES
+            self.nftables_enabled = False
+        else:
+            self.ip4tables_enabled = True
+            self.ipv4_supported_icmp_types = [ ]
+            self.ip6tables_enabled = True
+            self.ipv6_supported_icmp_types = [ ]
+            self.ebtables_enabled = True
+            self.ipset_enabled = True
+            self.ipset_supported_types = IPSET_TYPES
+            self.nftables_enabled = True
+
+    def get_all_io_objects_dict(self):
+        """
+        Returns a dict of dicts of all runtime config objects.
+        """
+        conf_dict = {}
+        conf_dict["ipsets"] = {_ipset: self.ipset.get_ipset(_ipset) for _ipset in self.ipset.get_ipsets()}
+        conf_dict["helpers"] = {helper: self.helper.get_helper(helper) for helper in self.helper.get_helpers()}
+        conf_dict["icmptypes"] = {icmptype: self.icmptype.get_icmptype(icmptype) for icmptype in self.icmptype.get_icmptypes()}
+        conf_dict["services"] = {service: self.service.get_service(service) for service in self.service.get_services()}
+        conf_dict["zones"] = {zone: self.zone.get_zone(zone) for zone in self.zone.get_zones()}
+        conf_dict["policies"] = {policy: self.policy.get_policy(policy) for policy in self.policy.get_policies_not_derived_from_zone()}
+
+        # The runtime might not actually support all the defined icmptypes.
+        # This is the case if ipv6 (ip6tables) is disabled. Unfortunately users
+        # disable IPv6 and also expect the IPv6 stuff to be silently ignored.
+        # This is problematic for defaults that include IPv6 stuff, e.g. policy
+        # 'allow-host-ipv6'. Use this to make a better decision about errors vs
+        # warnings.
+        #
+        conf_dict["icmptypes_unsupported"] = {}
+        for icmptype in (set(self.config.get_icmptypes()).difference(
+                         set(self.icmptype.get_icmptypes()))):
+            conf_dict["icmptypes_unsupported"][icmptype] = self.config.get_icmptype(icmptype)
+        # Some icmptypes support multiple families. Add those that are missing
+        # support for a subset of families.
+        for icmptype in (set(self.config.get_icmptypes()).intersection(
+                         set(self.icmptype.get_icmptypes()))):
+            if icmptype not in self.ipv4_supported_icmp_types or \
+               icmptype not in self.ipv6_supported_icmp_types:
+                conf_dict["icmptypes_unsupported"][icmptype] = copy.copy(self.config.get_icmptype(icmptype))
+                conf_dict["icmptypes_unsupported"][icmptype].destination = []
+                if icmptype not in self.ipv4_supported_icmp_types:
+                    conf_dict["icmptypes_unsupported"][icmptype].destination.append("ipv4")
+                if icmptype not in self.ipv6_supported_icmp_types:
+                    conf_dict["icmptypes_unsupported"][icmptype].destination.append("ipv6")
+
+        return conf_dict
+
+    def full_check_config(self, extra_io_objects: Dict[str, List[IO_Object]] = {}):
+        all_io_objects = self.get_all_io_objects_dict()
+        # mix in the extra objects
+        for type_key in extra_io_objects:
+            for obj in extra_io_objects[type_key]:
+                all_io_objects[type_key][obj.name] = obj
+
+        # we need to check in a well defined order because some io_objects will
+        # cross-check others
+        order = ["ipsets", "helpers", "icmptypes", "services", "zones", "policies"]
+        for io_obj_type in order:
+            io_objs = all_io_objects[io_obj_type]
+            for (name, io_obj) in io_objs.items():
+                io_obj.check_config_dict(io_obj.export_config_dict(), all_io_objects)
+
+    def _start_check_tables(self):
         # check if iptables, ip6tables and ebtables are usable, else disable
         if self.ip4tables_enabled and \
            "filter" not in self.ip4tables_backend.get_available_tables():
@@ -150,10 +210,9 @@ class Firewall(object):
         # is there at least support for ipv4 or ipv6
         if not self.ip4tables_enabled and not self.ip6tables_enabled \
            and not self.nftables_enabled:
-            log.fatal("No IPv4 and IPv6 firewall.")
-            sys.exit(1)
+            raise FirewallError(errors.UNKNOWN_ERROR, "No IPv4 and IPv6 firewall.")
 
-    def _start_check(self):
+    def _start_probe_backends(self):
         try:
             self.ipset_backend.set_list()
         except ValueError:
@@ -227,10 +286,7 @@ class Firewall(object):
             log.debug1("ebtables-restore is not supporting the --noflush "
                        "option, will therefore not be used")
 
-    def _start(self, reload=False, complete_reload=False):
-        # initialize firewall
-        default_zone = config.FALLBACK_ZONE
-
+    def _start_load_firewalld_conf(self):
         # load firewalld config
         log.debug1("Loading firewalld config file '%s'", config.FIREWALLD_CONF)
         try:
@@ -240,7 +296,7 @@ class Firewall(object):
             log.warning("Using fallback firewalld configuration settings.")
         else:
             if self._firewalld_conf.get("DefaultZone"):
-                default_zone = self._firewalld_conf.get("DefaultZone")
+                self._default_zone = self._firewalld_conf.get("DefaultZone")
 
             if self._firewalld_conf.get("CleanupOnExit"):
                 value = self._firewalld_conf.get("CleanupOnExit")
@@ -317,11 +373,7 @@ class Firewall(object):
 
         self.config.set_firewalld_conf(copy.deepcopy(self._firewalld_conf))
 
-        self._select_firewall_backend(self._firewall_backend)
-
-        if not self._offline:
-            self._start_check()
-
+    def _start_load_lockdown_whitelist(self):
         # load lockdown whitelist
         log.debug1("Loading lockdown whitelist")
         try:
@@ -337,64 +389,71 @@ class Firewall(object):
         # copy policies to config interface
         self.config.set_policies(copy.deepcopy(self.policies))
 
-        # load ipset files
-        self._loader(config.FIREWALLD_IPSETS, "ipset")
-        self._loader(config.ETC_FIREWALLD_IPSETS, "ipset")
+    def _start_load_stock_config(self):
+        self._loader_ipsets(config.FIREWALLD_IPSETS)
+        self._loader_icmptypes(config.FIREWALLD_ICMPTYPES)
+        self._loader_helpers(config.FIREWALLD_HELPERS)
+        self._loader_services(config.FIREWALLD_SERVICES)
+        self._loader_zones(config.FIREWALLD_ZONES)
+        self._loader_policies(config.FIREWALLD_POLICIES)
 
-        # load icmptype files
-        self._loader(config.FIREWALLD_ICMPTYPES, "icmptype")
-        self._loader(config.ETC_FIREWALLD_ICMPTYPES, "icmptype")
+    def _start_load_user_config(self):
+        self._loader_ipsets(config.ETC_FIREWALLD_IPSETS)
+        self._loader_icmptypes(config.ETC_FIREWALLD_ICMPTYPES)
+        self._loader_helpers(config.ETC_FIREWALLD_HELPERS)
+        self._loader_services(config.ETC_FIREWALLD_SERVICES)
+        self._loader_zones(config.ETC_FIREWALLD_ZONES)
+        self._loader_policies(config.ETC_FIREWALLD_POLICIES)
 
-        if len(self.icmptype.get_icmptypes()) == 0:
-            log.error("No icmptypes found.")
+    def _start_copy_config_to_runtime(self):
+        for _ipset in self.config.get_ipsets():
+            self.ipset.add_ipset(
+                    copy.deepcopy(self.config.get_ipset(_ipset)))
+        for icmptype in self.config.get_icmptypes():
+            self.icmptype.add_icmptype(
+                    copy.deepcopy(self.config.get_icmptype(icmptype)))
+        for helper in self.config.get_helpers():
+            self.helper.add_helper(
+                    copy.deepcopy(self.config.get_helper(helper)))
+        for service in self.config.get_services():
+            self.service.add_service(
+                    copy.deepcopy(self.config.get_service(service)))
+        for policy in self.config.get_policy_objects():
+            self.policy.add_policy(
+                    copy.deepcopy(self.config.get_policy_object(policy)))
 
-        # load helper files
-        self._loader(config.FIREWALLD_HELPERS, "helper")
-        self._loader(config.ETC_FIREWALLD_HELPERS, "helper")
+        self.direct.set_permanent_config(
+                copy.deepcopy(self.config.get_direct()))
 
-        # load service files
-        self._loader(config.FIREWALLD_SERVICES, "service")
-        self._loader(config.ETC_FIREWALLD_SERVICES, "service")
+        # copy combined permanent zones to runtime
+        # zones with a '/' in the name will be combined into one runtime zone
+        combined_zones = {}
+        for zone in self.config.get_zones():
+            z_obj = self.config.get_zone(zone)
+            if '/' not in z_obj.name:
+                self.zone.add_zone(
+                        copy.deepcopy(self.config.get_zone(zone)))
+                continue
 
-        if len(self.service.get_services()) == 0:
-            log.error("No services found.")
+            combined_name = os.path.basename(z_obj.path)
+            if combined_name not in combined_zones:
+                combined_zone = Zone()
+                combined_zone.name = combined_name
+                combined_zone.check_name(combined_zone.name)
+                combined_zone.path = z_obj.path
+                combined_zone.default = False
+                combined_zone.forward = False # see note in zone_reader()
 
-        # load zone files
-        self._loader(config.FIREWALLD_ZONES, "zone")
-        self._loader(config.ETC_FIREWALLD_ZONES, "zone")
+                combined_zones[combined_name] = combined_zone
 
-        if len(self.zone.get_zones()) == 0:
-            log.fatal("No zones found.")
-            sys.exit(1)
+            log.debug1("Combining zone '%s' using '%s%s%s'",
+                       combined_name, z_obj.path, os.sep, z_obj.filename)
+            combined_zones[combined_name].combine(z_obj)
 
-        # load policy files
-        self._loader(config.FIREWALLD_POLICIES, "policy")
-        self._loader(config.ETC_FIREWALLD_POLICIES, "policy")
-
-        # check minimum required zones
-        error = False
-        for z in [ "block", "drop", "trusted" ]:
-            if z not in self.zone.get_zones():
-                log.fatal("Zone '%s' is not available.", z)
-                error = True
-        if error:
-            sys.exit(1)
-
-        # check if default_zone is a valid zone
-        if default_zone not in self.zone.get_zones():
-            if "public" in self.zone.get_zones():
-                zone = "public"
-            elif "external" in self.zone.get_zones():
-                zone = "external"
-            else:
-                zone = "block" # block is a base zone, therefore it has to exist
-
-            log.error("Default zone '%s' is not valid. Using '%s'.",
-                      default_zone, zone)
-            default_zone = zone
-        else:
-            log.debug1("Using default zone '%s'", default_zone)
+        for zone in combined_zones:
+            self.zone.add_zone(combined_zones[zone])
 
+    def _start_load_direct_rules(self):
         # load direct rules
         obj = Direct(config.FIREWALLD_DIRECT)
         if os.path.exists(config.FIREWALLD_DIRECT):
@@ -405,25 +464,11 @@ class Firewall(object):
             except Exception as msg:
                 log.error("Failed to load direct rules file '%s': %s",
                           config.FIREWALLD_DIRECT, msg)
-        self.direct.set_permanent_config(obj)
-        self.config.set_direct(copy.deepcopy(obj))
+        self.config.set_direct(obj)
 
-        self._default_zone = self.check_zone(default_zone)
-
-        if self._offline:
-            return
-
-        # check if needed tables are there
-        self._check_tables()
-
-        if log.getDebugLogLevel() > 0:
-            # get time before flushing and applying
-            tm1 = time.time()
-
-        # Start transaction
+    def _start_apply_objects(self, reload=False, complete_reload=False):
         transaction = FirewallTransaction(self)
 
-        # flush rules
         self.flush(use_transaction=transaction)
 
         # If modules need to be unloaded in complete reload or if there are
@@ -451,29 +496,24 @@ class Firewall(object):
             log.debug1("Applying ipsets")
             self.ipset.apply_ipsets()
 
-        # Start or continue with transaction
-
-        # apply default rules
         log.debug1("Applying default rule set")
         self.apply_default_rules(use_transaction=transaction)
 
-        # apply settings for loaded zones
         log.debug1("Applying used zones")
         self.zone.apply_zones(use_transaction=transaction)
 
         self.zone.change_default_zone(None, self._default_zone,
                                       use_transaction=transaction)
 
-        # apply policies
         log.debug1("Applying used policies")
         self.policy.apply_policies(use_transaction=transaction)
 
-        # Execute transaction
         transaction.execute(True)
-
-        # Start new transaction for direct rules
         transaction.clear()
 
+    def _start_apply_direct_rules(self):
+        transaction = FirewallTransaction(self)
+
         # apply direct chains, rules and passthrough rules
         if self.direct.has_configuration():
             log.debug1("Applying direct chains rules and passthrough rules")
@@ -489,179 +529,241 @@ class Firewall(object):
             except Exception:
                 raise
 
-        del transaction
+        transaction.execute(True)
+        transaction.clear()
+
+    def _start_check(self):
+        # check minimum required zones
+        for z in [ "block", "drop", "trusted" ]:
+            if z not in self.zone.get_zones():
+                raise FirewallError(errors.INVALID_ZONE, "Zone '{}' is not available.".format(z))
+
+        # check if default_zone is a valid zone
+        if self._default_zone not in self.zone.get_zones():
+            if "public" in self.zone.get_zones():
+                zone = "public"
+            elif "external" in self.zone.get_zones():
+                zone = "external"
+            else:
+                zone = "block" # block is a base zone, therefore it has to exist
+
+            log.error("Default zone '%s' is not valid. Using '%s'.",
+                      self._default_zone, zone)
+            self._default_zone = zone
+        else:
+            log.debug1("Using default zone '%s'", self._default_zone)
+
+        if not self._offline:
+            self.full_check_config()
+            self._start_check_tables()
+
+            # check our desired backend is actually available
+            if self._firewall_backend == "iptables":
+                backend_to_check = "ip4tables" # ip6tables is always optional
+            else:
+                backend_to_check = self._firewall_backend
+            if not self.is_backend_enabled(backend_to_check):
+                raise FirewallError(errors.UNKNOWN_ERROR,
+                        "Firewall backend '{}' is not available.".format(
+                            self._firewall_backend))
+
+    def _start(self, reload=False, complete_reload=False):
+        self._start_load_firewalld_conf()
+        self._start_load_lockdown_whitelist()
+
+        self._select_firewall_backend(self._firewall_backend)
+
+        if not self._offline:
+            self._start_probe_backends()
+
+        self._start_load_stock_config()
+        self._start_load_user_config()
+        self._start_load_direct_rules()
+        self._start_copy_config_to_runtime()
+
+        self._start_check()
+
+        if self._offline:
+            return
+
+        if log.getDebugLogLevel() > 0:
+            # get time before flushing and applying
+            tm1 = time.time()
+
+        self._start_apply_objects(reload=reload, complete_reload=complete_reload)
+        self._start_apply_direct_rules()
 
         if log.getDebugLogLevel() > 1:
             # get time after flushing and applying
             tm2 = time.time()
             log.debug2("Flushing and applying took %f seconds" % (tm2 - tm1))
 
+    def _start_failsafe(self, reload=False, complete_reload=False):
+        """
+        This is basically _start() with at least the following differences:
+            - built-in defaults for firewalld.conf
+            - no lockdown list
+            - no user config (/etc/firewalld)
+            - no direct rules
+        """
+        self.cleanup()
+        self._firewalld_conf.set_defaults()
+        self.config.set_firewalld_conf(copy.deepcopy(self._firewalld_conf))
+
+        self._select_firewall_backend(self._firewall_backend)
+
+        if not self._offline:
+            self._start_probe_backends()
+
+        self._start_load_stock_config()
+        self._start_copy_config_to_runtime()
+        self._start_check()
+
+        if self._offline:
+            return
+
+        self._start_apply_objects(reload=reload, complete_reload=complete_reload)
+
     def start(self):
         try:
             self._start()
-        except Exception:
-            self._state = "FAILED"
-            self.set_policy("ACCEPT")
-            raise
+        except Exception as original_ex:
+            log.error("Failed to load user configuration. Falling back to "
+                      "full stock configuration.")
+            try:
+                self._start_failsafe()
+                self._state = "FAILED"
+                self.set_policy("ACCEPT")
+            except Exception:
+                log.error(original_ex)
+                log.error("Failed to load full stock configuration. This likely "
+                          "indicates a system level issue, e.g. the firewall "
+                          "backend (nftables, iptables) is broken. "
+                          "All hope is lost. Exiting.")
+                try:
+                    self.flush()
+                except Exception:
+                    pass
+                sys.exit(errors.UNKNOWN_ERROR)
+            # propagate the original exception that caused us to enter failed
+            # state.
+            raise original_ex
         else:
             self._state = "RUNNING"
             self.set_policy("ACCEPT")
 
-    def _loader(self, path, reader_type, combine=False):
-        # combine: several zone files are getting combined into one obj
+    def _loader_config_file_generator(self, path):
         if not os.path.isdir(path):
             return
 
-        if combine:
-            if path.startswith(config.ETC_FIREWALLD) and reader_type == "zone":
-                combined_zone = Zone()
-                combined_zone.name = os.path.basename(path)
-                combined_zone.check_name(combined_zone.name)
-                combined_zone.path = path
-                combined_zone.default = False
-            else:
-                combine = False
+        for filename in sorted(os.listdir(path)):
+            if not filename.endswith(".xml"):
+                continue
+            yield filename
+
+    def _loader_services(self, path):
+        for filename in self._loader_config_file_generator(path):
+            log.debug1("Loading service file '%s%s%s'", path, os.sep, filename)
+
+            obj = service_reader(filename, path)
+            if obj.name in self.config.get_services():
+                orig_obj = self.config.get_service(obj.name)
+                log.debug1("Overrides '%s%s%s'",
+                           orig_obj.path, os.sep, orig_obj.filename)
+            elif obj.path.startswith(config.ETC_FIREWALLD):
+                obj.default = True
+
+            self.config.add_service(obj)
+
+    def _loader_ipsets(self, path):
+        for filename in self._loader_config_file_generator(path):
+            log.debug1("Loading ipset file '%s%s%s'", path, os.sep, filename)
+
+            obj = ipset_reader(filename, path)
+            if obj.name in self.config.get_ipsets():
+                orig_obj = self.config.get_ipset(obj.name)
+                log.debug1("Overrides '%s%s%s'",
+                           orig_obj.path, os.sep, orig_obj.filename)
+            elif obj.path.startswith(config.ETC_FIREWALLD):
+                obj.default = True
+
+            self.config.add_ipset(obj)
+
+    def _loader_helpers(self, path):
+        for filename in self._loader_config_file_generator(path):
+            log.debug1("Loading helper file '%s%s%s'", path, os.sep, filename)
+
+            obj = helper_reader(filename, path)
+            if obj.name in self.config.get_helpers():
+                orig_obj = self.config.get_helper(obj.name)
+                log.debug1("Overrides '%s%s%s'",
+                           orig_obj.path, os.sep, orig_obj.filename)
+            elif obj.path.startswith(config.ETC_FIREWALLD):
+                obj.default = True
+
+            self.config.add_helper(obj)
+
+    def _loader_policies(self, path):
+        for filename in self._loader_config_file_generator(path):
+            log.debug1("Loading policy file '%s%s%s'", path, os.sep, filename)
+
+            obj = policy_reader(filename, path)
+            if obj.name in self.config.get_policy_objects():
+                orig_obj = self.config.get_policy_object(obj.name)
+                log.debug1("Overrides '%s%s%s'",
+                           orig_obj.path, os.sep, orig_obj.filename)
+            elif obj.path.startswith(config.ETC_FIREWALLD):
+                obj.default = True
+
+            self.config.add_policy_object(obj)
+
+    def _loader_icmptypes(self, path):
+        for filename in self._loader_config_file_generator(path):
+            log.debug1("Loading icmptype file '%s%s%s'", path, os.sep, filename)
+
+            obj = icmptype_reader(filename, path)
+            if obj.name in self.config.get_icmptypes():
+                orig_obj = self.config.get_icmptype(obj.name)
+                log.debug1("Overrides '%s%s%s'",
+                           orig_obj.path, os.sep, orig_obj.filename)
+            elif obj.path.startswith(config.ETC_FIREWALLD):
+                obj.default = True
+
+            self.config.add_icmptype(obj)
+
+    def _loader_zones(self, path, combine=False):
+        if not os.path.isdir(path):
+            return
 
         for filename in sorted(os.listdir(path)):
             if not filename.endswith(".xml"):
                 if path.startswith(config.ETC_FIREWALLD) and \
-                        reader_type == "zone" and \
                         os.path.isdir("%s/%s" % (path, filename)):
-                    self._loader("%s/%s" % (path, filename), reader_type,
-                                 combine=True)
+                    # Combined zones are added to permanent config
+                    # individually. They're coalesced into one object when
+                    # added to the runtime
+                    self._loader_zones("%s/%s" % (path, filename), combine=True)
                 continue
 
             name = "%s/%s" % (path, filename)
-            log.debug1("Loading %s file '%s'", reader_type, name)
-            try:
-                if reader_type == "icmptype":
-                    obj = icmptype_reader(filename, path)
-                    if obj.name in self.icmptype.get_icmptypes():
-                        orig_obj = self.icmptype.get_icmptype(obj.name)
-                        log.debug1("  Overloads %s '%s' ('%s/%s')", reader_type,
-                                   orig_obj.name, orig_obj.path,
-                                   orig_obj.filename)
-                        self.icmptype.remove_icmptype(orig_obj.name)
-                    elif obj.path.startswith(config.ETC_FIREWALLD):
-                        obj.default = True
-                    try:
-                        self.icmptype.add_icmptype(obj)
-                    except FirewallError as error:
-                        log.info1("%s: %s, ignoring for run-time." % \
-                                    (obj.name, str(error)))
-                    # add a deep copy to the configuration interface
-                    self.config.add_icmptype(copy.deepcopy(obj))
-                elif reader_type == "service":
-                    obj = service_reader(filename, path)
-                    if obj.name in self.service.get_services():
-                        orig_obj = self.service.get_service(obj.name)
-                        log.debug1("  Overloads %s '%s' ('%s/%s')", reader_type,
-                                   orig_obj.name, orig_obj.path,
-                                   orig_obj.filename)
-                        self.service.remove_service(orig_obj.name)
-                    elif obj.path.startswith(config.ETC_FIREWALLD):
-                        obj.default = True
-                    self.service.add_service(obj)
-                    # add a deep copy to the configuration interface
-                    self.config.add_service(copy.deepcopy(obj))
-                elif reader_type == "zone":
-                    obj = zone_reader(filename, path, no_check_name=combine)
-                    if combine:
-                        # Change name for permanent configuration
-                        obj.name = "%s/%s" % (
-                            os.path.basename(path),
-                            os.path.basename(filename)[0:-4])
-                        obj.check_name(obj.name)
-                    # Copy object before combine
-                    config_obj = copy.deepcopy(obj)
-                    if obj.name in self.zone.get_zones():
-                        orig_obj = self.zone.get_zone(obj.name)
-                        self.zone.remove_zone(orig_obj.name)
-                        if orig_obj.combined:
-                            log.debug1("  Combining %s '%s' ('%s/%s')",
-                                        reader_type, obj.name,
-                                        path, filename)
-                            obj.combine(orig_obj)
-                        else:
-                            log.debug1("  Overloads %s '%s' ('%s/%s')",
-                                       reader_type,
-                                       orig_obj.name, orig_obj.path,
-                                       orig_obj.filename)
-                    elif obj.path.startswith(config.ETC_FIREWALLD):
-                        obj.default = True
-                        config_obj.default = True
-                    self.config.add_zone(config_obj)
-                    if combine:
-                        log.debug1("  Combining %s '%s' ('%s/%s')",
-                                   reader_type, combined_zone.name,
-                                   path, filename)
-                        combined_zone.combine(obj)
-                    else:
-                        self.zone.add_zone(obj)
-                elif reader_type == "ipset":
-                    obj = ipset_reader(filename, path)
-                    if obj.name in self.ipset.get_ipsets():
-                        orig_obj = self.ipset.get_ipset(obj.name)
-                        log.debug1("  Overloads %s '%s' ('%s/%s')", reader_type,
-                                   orig_obj.name, orig_obj.path,
-                                   orig_obj.filename)
-                        self.ipset.remove_ipset(orig_obj.name)
-                    elif obj.path.startswith(config.ETC_FIREWALLD):
-                        obj.default = True
-                    try:
-                        self.ipset.add_ipset(obj)
-                    except FirewallError as error:
-                        log.warning("%s: %s, ignoring for run-time." % \
-                                    (obj.name, str(error)))
-                    # add a deep copy to the configuration interface
-                    self.config.add_ipset(copy.deepcopy(obj))
-                elif reader_type == "helper":
-                    obj = helper_reader(filename, path)
-                    if obj.name in self.helper.get_helpers():
-                        orig_obj = self.helper.get_helper(obj.name)
-                        log.debug1("  Overloads %s '%s' ('%s/%s')", reader_type,
-                                   orig_obj.name, orig_obj.path,
-                                   orig_obj.filename)
-                        self.helper.remove_helper(orig_obj.name)
-                    elif obj.path.startswith(config.ETC_FIREWALLD):
-                        obj.default = True
-                    self.helper.add_helper(obj)
-                    # add a deep copy to the configuration interface
-                    self.config.add_helper(copy.deepcopy(obj))
-                elif reader_type == "policy":
-                    obj = policy_reader(filename, path)
-                    if obj.name in self.policy.get_policies():
-                        orig_obj = self.policy.get_policy(obj.name)
-                        log.debug1("  Overloads %s '%s' ('%s/%s')", reader_type,
-                                   orig_obj.name, orig_obj.path,
-                                   orig_obj.filename)
-                        self.policy.remove_policy(orig_obj.name)
-                    elif obj.path.startswith(config.ETC_FIREWALLD):
-                        obj.default = True
-                    self.policy.add_policy(obj)
-                    # add a deep copy to the configuration interface
-                    self.config.add_policy_object(copy.deepcopy(obj))
-                else:
-                    log.fatal("Unknown reader type %s", reader_type)
-            except FirewallError as msg:
-                log.error("Failed to load %s file '%s': %s", reader_type,
-                          name, msg)
-            except Exception:
-                log.error("Failed to load %s file '%s':", reader_type, name)
-                log.exception()
+            log.debug1("Loading zone file '%s'", name)
 
-        if combine and combined_zone.combined:
-            if combined_zone.name in self.zone.get_zones():
-                orig_obj = self.zone.get_zone(combined_zone.name)
-                log.debug1("  Overloading and deactivating %s '%s' ('%s/%s')",
-                           reader_type, orig_obj.name, orig_obj.path,
-                           orig_obj.filename)
-                try:
-                    self.zone.remove_zone(combined_zone.name)
-                except Exception:
-                    pass
-                self.config.forget_zone(combined_zone.name)
-            self.zone.add_zone(combined_zone)
+            obj = zone_reader(filename, path, no_check_name=combine)
+            if combine:
+                # Change name for permanent configuration
+                obj.name = "%s/%s" % (
+                    os.path.basename(path),
+                    os.path.basename(filename)[0:-4])
+                obj.check_name(obj.name)
+
+            if obj.name in self.config.get_zones():
+                orig_obj = self.config.get_zone(obj.name)
+                log.debug1("Overrides '%s%s%s'",
+                           orig_obj.path, os.sep, orig_obj.filename)
+            elif obj.path.startswith(config.ETC_FIREWALLD):
+                obj.default = True
+
+            self.config.add_zone(obj)
 
     def cleanup(self):
         self.icmptype.cleanup()
@@ -980,6 +1082,9 @@ class Firewall(object):
     # RELOAD
 
     def reload(self, stop=False):
+        # we're about to load the on-disk config, so verify it's sane.
+        check_on_disk_config(self)
+
         _panic = self._panic
 
         # must stash this. The value may change after _start()
@@ -989,7 +1094,7 @@ class Firewall(object):
             # save zone interfaces
             _zone_interfaces = { }
             for zone in self.zone.get_zones():
-                _zone_interfaces[zone] = self.zone.get_settings(zone)["interfaces"]
+                _zone_interfaces[zone] = self.zone.get_zone(zone).interfaces
             # save direct config
             _direct_config = self.direct.get_runtime_config()
             _old_dz = self.get_default_zone()
@@ -1033,8 +1138,8 @@ class Firewall(object):
                     _zone_interfaces[_new_dz] = { }
                 # default zone changed. Move interfaces from old default zone to
                 # the new one.
-                for iface, settings in list(_zone_interfaces[_old_dz].items()):
-                    if settings["__default__"]:
+                for iface in _zone_interfaces[_old_dz]:
+                    if iface in self._default_zone_interfaces:
                         # move only those that were added to default zone
                         # (not those that were added to specific zone same as
                         # default)
@@ -1047,8 +1152,7 @@ class Firewall(object):
                 if zone in _zone_interfaces:
 
                     for interface_id in _zone_interfaces[zone]:
-                        self.zone.change_zone_of_interface(zone, interface_id,
-                                                           _zone_interfaces[zone][interface_id]["sender"])
+                        self.zone.change_zone_of_interface(zone, interface_id)
 
                     del _zone_interfaces[zone]
                 else:
@@ -1155,13 +1259,15 @@ class Firewall(object):
             self._firewalld_conf.set("DefaultZone", _zone)
             self._firewalld_conf.write()
 
+            if self._offline:
+                return
+
             # remove old default zone from ZONES and add new default zone
             self.zone.change_default_zone(_old_dz, _zone)
 
             # Move interfaces from old default zone to the new one.
-            _old_dz_settings = self.zone.get_settings(_old_dz)
-            for iface, settings in list(_old_dz_settings["interfaces"].items()):
-                if settings["__default__"]:
+            for iface in self.zone.get_zone(_old_dz).interfaces:
+                if iface in self._default_zone_interfaces:
                     # move only those that were added to default zone
                     # (not those that were added to specific zone same as default)
                     self.zone.change_zone_of_interface("", iface)
@@ -1183,6 +1289,11 @@ class Firewall(object):
         return combined
 
     def get_added_and_removed_settings(self, old_settings, new_settings):
+        # normalize rich rules, zones and policies use a different key
+        for rich_key in ["rich_rules", "rules_str"]:
+            if rich_key in new_settings:
+                new_settings[rich_key] = [str(Rich_Rule(rule_str=rule_str)) for rule_str in new_settings[rich_key]]
+
         add_settings = {}
         remove_settings = {}
         for key in (set(old_settings.keys()) | set(new_settings.keys())):
