@@ -28,7 +28,7 @@ from firewall.functions import tempFile, readfile, splitArgs, check_mac, portStr
                                check_single_address, check_address, normalizeIP6
 from firewall import config
 from firewall.errors import FirewallError, INVALID_PASSTHROUGH, INVALID_RULE, UNKNOWN_ERROR, INVALID_ADDR
-from firewall.core.rich import Rich_Accept, Rich_Reject, Rich_Drop, Rich_Mark, \
+from firewall.core.rich import Rich_Accept, Rich_Reject, Rich_Drop, Rich_Mark, Rich_NFLog, \
                                Rich_Masquerade, Rich_ForwardPort, Rich_IcmpBlock, Rich_Tcp_Mss_Clamp
 from firewall.core.base import DEFAULT_ZONE_TARGET
 import string
@@ -676,7 +676,13 @@ class ip4tables(object):
                 default_rules["nat"].append("-A %s -j %s_direct" % (chain, chain))
                 self.our_chains["nat"].add("%s_direct" % chain)
 
-                if chain in [ "PREROUTING", "POSTROUTING" ]:
+                if chain in ["OUTPUT"]:
+                    # nat, output does not have zone dispatch
+                    for dispatch_suffix in ["POLICIES_pre", "POLICIES_post"]:
+                        default_rules["nat"].append("-N %s_%s" % (chain, dispatch_suffix))
+                        self.our_chains["nat"].update(set(["%s_%s" % (chain, dispatch_suffix)]))
+                        default_rules["nat"].append("-A %s -j %s_%s" % (chain, chain, dispatch_suffix))
+                else:
                     for dispatch_suffix in ["POLICIES_pre", "ZONES", "POLICIES_post"]:
                         default_rules["nat"].append("-N %s_%s" % (chain, dispatch_suffix))
                         self.our_chains["nat"].update(set(["%s_%s" % (chain, dispatch_suffix)]))
@@ -951,11 +957,11 @@ class ip4tables(object):
                 if target in [DEFAULT_ZONE_TARGET, "REJECT", "%%REJECT%%" ]:
                     rules.append([ add_del_rule, _policy, "-t", table, "%%LOGTYPE%%",
                                    "-j", "LOG", "--log-prefix",
-                                   "\"%s_REJECT: \"" % _policy ])
+                                   "%s_REJECT: " % _policy ])
                 if target == "DROP":
                     rules.append([ add_del_rule, _policy, "-t", table, "%%LOGTYPE%%",
                                    "-j", "LOG", "--log-prefix",
-                                   "\"%s_DROP: \"" % _policy ])
+                                   "%s_DROP: " % _policy ])
 
         if table == "filter" and \
            target in [DEFAULT_ZONE_TARGET, "ACCEPT", "REJECT", "%%REJECT%%", "DROP" ]:
@@ -1024,11 +1030,20 @@ class ip4tables(object):
         chain_suffix = self._rich_rule_chain_suffix_from_log(rich_rule)
         rule = ["-t", table, add_del, "%s_%s" % (_policy, chain_suffix)]
         rule += self._rich_rule_priority_fragment(rich_rule)
-        rule += rule_fragment + [ "-j", "LOG" ]
-        if rich_rule.log.prefix:
-            rule += [ "--log-prefix", "'%s'" % rich_rule.log.prefix ]
-        if rich_rule.log.level:
-            rule += [ "--log-level", "%s" % rich_rule.log.level ]
+        if type(rich_rule.log) == Rich_NFLog:
+            rule += rule_fragment + [ "-j", "NFLOG" ]
+            if rich_rule.log.group:
+                rule += [ "--nflog-group", rich_rule.log.group ]
+            if rich_rule.log.prefix:
+                rule += [ "--nflog-prefix", "%s" % rich_rule.log.prefix ]
+            if rich_rule.log.threshold:
+                rule += [ "--nflog-threshold", rich_rule.log.threshold ]
+        else:
+            rule += rule_fragment + [ "-j", "LOG" ]
+            if rich_rule.log.prefix:
+                rule += [ "--log-prefix", "%s" % rich_rule.log.prefix ]
+            if rich_rule.log.level:
+                rule += [ "--log-level", "%s" % rich_rule.log.level ]
         rule += self._rule_limit(rich_rule.log.limit)
 
         return rule
@@ -1384,7 +1399,7 @@ class ip4tables(object):
                 rules.append([ add_del, final_chain, "-t", table ]
                              + rule_fragment +
                              [ "%%LOGTYPE%%", "-j", "LOG",
-                               "--log-prefix", "\"%s_ICMP_BLOCK: \"" % policy ])
+                               "--log-prefix", "%s_ICMP_BLOCK: " % policy ])
             rules.append([ add_del, final_chain, "-t", table ]
                          + rule_fragment +
                          [ "-j", final_target ])
@@ -1410,7 +1425,7 @@ class ip4tables(object):
                 rule = rule + [ "-t", table, "-p", "%%ICMP%%",
                               "%%LOGTYPE%%",
                               "-j", "LOG", "--log-prefix",
-                              "\"%s_ICMP_BLOCK: \"" % _policy ]
+                              "%s_ICMP_BLOCK: " % _policy ]
                 rules.append(rule)
                 rule_idx += 1
         else:
@@ -1468,16 +1483,16 @@ class ip6tables(ip4tables):
 
     def build_rfc3964_ipv4_rules(self):
         daddr_list = [
-                     "::0.0.0.0/96", # IPv4 compatible
-                     "::ffff:0.0.0.0/96", # IPv4 mapped
-                     "2002:0000::/24", # 0.0.0.0/8 (the system has no address assigned yet)
-                     "2002:0a00::/24", # 10.0.0.0/8 (private)
-                     "2002:7f00::/24", # 127.0.0.0/8 (loopback)
-                     "2002:ac10::/28", # 172.16.0.0/12 (private)
-                     "2002:c0a8::/32", # 192.168.0.0/16 (private)
-                     "2002:a9fe::/32", # 169.254.0.0/16 (IANA Assigned DHCP link-local)
-                     "2002:e000::/19", # 224.0.0.0/4 (multicast), 240.0.0.0/4 (reserved and broadcast)
-                     ]
+            "::0.0.0.0/96", # IPv4 compatible
+            "::ffff:0.0.0.0/96", # IPv4 mapped
+            "2002:0000::/24", # 0.0.0.0/8 (the system has no address assigned yet)
+            "2002:0a00::/24", # 10.0.0.0/8 (private)
+            "2002:7f00::/24", # 127.0.0.0/8 (loopback)
+            "2002:ac10::/28", # 172.16.0.0/12 (private)
+            "2002:c0a8::/32", # 192.168.0.0/16 (private)
+            "2002:a9fe::/32", # 169.254.0.0/16 (IANA Assigned DHCP link-local)
+            "2002:e000::/19", # 224.0.0.0/4 (multicast), 240.0.0.0/4 (reserved and broadcast)
+        ]
 
         chain_name = "RFC3964_IPv4"
         self.our_chains["filter"].add(chain_name)
@@ -1491,7 +1506,7 @@ class ip6tables(ip4tables):
             if self._fw._log_denied in ["unicast", "all"]:
                 rules.append(["-t", "filter", "-I", chain_name,
                               "-d", daddr, "-j", "LOG",
-                              "--log-prefix", "\"RFC3964_IPv4_REJECT: \""])
+                              "--log-prefix", "RFC3964_IPv4_REJECT: "])
 
         # Inject into FORWARD and OUTPUT chains
         rules.append(["-t", "filter", "-I", "OUTPUT", "4",
